; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     x87 Hardware Emulator                           *
; *                                                     *
; *     Copyright (C) 2000-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   EM87

             INCLUDE SE.ASM

             .386
             .387

        if DPMIVersion

             EXTRN   __AHIncr:ABS

        endif

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

             PUBLIC  E087_ENTRY
             PUBLIC  E087_SHORTCUT

;-------------------------------------------------------------------------------
; E087_ENTRY:
;
; This interrupt routine is used to translate the original, emulated through
; interrupts 34-3D, floating point code into real FPU instructions.
;
; Interrupt 34..3B emulate FPU opcodes D8..DF
; Interrupt 3D emulates a stand-alone FWAIT
; Interrupt 3C is used to emulate opcodes with segment overrides
;
; From Ralf Brown:
;
; The generated code is CD 3C xy mm ....
; where xy is a modified ESC instruction and mm is the modR/M byte.
;
; The xy byte appears to be encoded as
;   s s 0 1 1 x x x  or  s s 0 0 0 x x x
;
; where      "ss" specifies the segment override:
;   00 -> DS: [3Eh (001 11 110)]
;   01 -> SS: [36h (001 10 110)]
;   10 -> CS: [2Eh (001 01 110)]
;   11 -> ES: [26h (001 00 110)]
;-------------------------------------------------------------------------------
E087_ENTRY   PROC    NEAR

             push    bp
             mov     bp, sp
             push    ax
             push    si
             push    ds
             test    byte ptr [bp + 7], 02h  ; pushed Interrupt flag
             jz      @@1

             sti

        if DPMIVersion

@@1:         mov     si, [bp + 2]
             mov     ax, [bp + 4]
             add     ax, __AHIncr
             mov     ds, ax

        else

@@1:         lds     si, [bp + 2]            ; load interrupt return address

        endif

             dec     si                      ; back to get interrupt number
             mov     ax, [si]                ; get interrupt no[AL]/first byte[AH]
             dec     si                      ; back to interrupt opcode
             sub     al, 34h                 ; emulator uses 34..3D
             cmp     al, 09h
             ja      @@4                     ; not an emulator call

             je      @@5                     ; 3D: standalone wait

             cmp     al, 8                   ; 3C: segment override
             je      @@6

             add     al, 0D8h                ; convert back to D8..DF

@@2:         mov     ah, al                  ; AH will end up at higher address
             mov     al, 90h                 ; NOP

@@3:         mov     [si], ax
             mov     bp, sp
             mov     [bp + 8], si            ; return to NOP

@@4:         pop     ds
             pop     si
             pop     ax
             pop     bp

E087_SHORTCUT:

             iret

@@5:         mov     ax, 9B90h               ; NOP + FWAIT
             jmp     @@3

@@6:         test    ah, 20h                 ; 0010 0000 not valid, NOP
             jnz     @@4

             mov     al, ah                  ; duplicate first byte
             and     ax, 07C0h               ; 0000 0fgh ab00 0000
             shr     al, 3                   ;           000a b000
             xor     al, 18h
             add     ax, 0D826h              ; base + "seges"
             mov     [si + 2], ah
             jmp     @@2

E087_ENTRY   ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END