; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Exponentiation                             *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FEXP

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   RealAdd      : NEAR
             EXTRN   RealSub      : NEAR
             EXTRN   RealMul      : NEAR
             EXTRN   RealFloat    : NEAR
             EXTRN   RealTrunc    : NEAR
             EXTRN   RealSqr      : NEAR
             EXTRN   RealDivRev   : NEAR
             EXTRN   RealMulNoChk : NEAR
             EXTRN   RealReduceMP : NEAR
             EXTRN   ROverflow    : NEAR
             EXTRN   RealMulF     : NEAR
             EXTRN   RealMulFNoChk: NEAR

; Publics

             PUBLIC  RExp

;-------------------------------------------------------------------------------
; Routines RExp, RPower2, and RPower10 exponentiate their argument x to base e,
; base 2, and base 10, respectively. All routines do the exponentiation of an
; argument x by calculating e^g * 2^n, where g is in -0.5*ln(2)..0.5*ln(2).
; While 2^n is a simple scaling operation, e^g is computed using a rational
; approximation of the Pade type. The computation of i and g differs for the
; three functions. For RPower2, n is computed as n = Round (x), and g as g =
; (x - n) * ln (2). For RExp, the computation proceeds as follows: n = Round
; (x/ln(2)), g = x - ln(2)*n, where ln(2) is represented by two constants c1
; and c2 that provide more than 40 bits of precision, so that the actual
; computation becomes x - c1*n - c2*n. This way, loss of accuracy due to the
; argument reduction is prevented. The computation of RPower10 is similar to
; that of the RExp function n = Round (x/log10(2)), g = x - log10(2) * n by
; the multiple precision process described above. Finally, g = g * ln(10).
; Actually, these different argument reduction schemes are all derived from
; the same process by elimination of unnecessary steps: n = Round (x/logB (2)),
; g = (x - n*logB(2)) * ln (B), where B is the base for exponentation.
;
; If the result of the exponentiation routines underflows the floating point
; format, zero is returned. If the computation results in overflow, error 205
; is returned through the error handler.
;
; GPZ := (2.001347076967e1*g^2+8.408086858319e2)*g
; QZ  := (g^2+1.801617224813e2)*g^2+1.681617371664e3
;
; e^g := 2 * (0.5 + GPZ / (QZ - GPZ))
;
; This approximation has a theoretical maximum relative error of 2.98e-14.
; Maximum observed error when evaluated in REAL arithmetic is 1.74e-12.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  2^argument, e^argument, 10^argument depending on function
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RExp_Ext     PROC    FAR

$argumnt_err:pop     bp                ; restore caller's frame pointer
             test    di, di            ; argument negative ?
             js      $exp_zero         ; yes, result = 0 to machine precision

             jmp     $exp_overfl       ; no, overflow, signal error

$exp_zero:   xor     ax, ax            ; load
             xor     bx, bx            ;  a
             xor     dx, dx            ;   zero
             ret                       ; done

RExp_Ext     ENDP

             INCLUDE ALIGN.ASM

RExp         PROC    FAR

             push    bp                ; save frame pointer
             mov     bp, offset PowECst; pointer to table of constants

$power_10:   push    dx                ; save
             push    bx                ;  argument x
             push    ax                ;   on stack
             mov     cx, cs:[bp + 12]  ; load
             mov     si, cs:[bp + 14]  ;  constant
             mov     di, cs:[bp + 16]  ;   1/lg(2)
             call    RealMulFNochk     ; compute y := x/lg(2) (DI:SI:CX <> 0)

             pop     cx                ; get
             pop     si                ;  back
             pop     di                ;   x
             cmp     al, 88h           ; abs (y) < 128 ?
             jnb     $argumnt_err      ; no, underflow or overflow

             push    cx                ; save lsw of x
             mov     ch, -1            ; signal rounding
             call    RealTrunc         ; n = round (y)

             pop     cx                ; get back lsw of x
             push    ax                ; save n
             push    cx                ; save lsw of x
             call    RealFloat         ; convert n to floating-point

             pop     cx                ; get back lsw of x
             call    RealReduceMP      ; compute t = x - n*c1 - n*c2

$approx:     push    dx                ; save
             push    bx                ;  g on
             push    ax                ;   stack
             call    RealSqr           ; compute z := g^2

             pop     cx                ; get
             pop     si                ;  g from
             pop     di                ;   stack
             push    dx                ; save
             push    bx                ;  z on
             push    ax                ;   stack
             push    di                ; save
             push    si                ;  g on
             push    cx                ;   stack
             mov     cx, 01A85h        ; load real
             mov     si, 09690h        ;  constant
             mov     di, 0201Bh        ;   2.001347076967e1
             call    RealMulFNoChk     ; multiply by z  (DI:SI:CX <> 0)

             mov     cx, 0388Ah        ; load
             mov     si, 0C182h        ;  real constant
             mov     di, 05233h        ;   8.408086858319e2
             call    RealAdd           ; make sum

             pop     cx                ; get
             pop     si                ;  g from
             pop     di                ;   stack
             call    RealMulF          ; compute GPZ

             pop     cx                ; get
             pop     si                ;  z from
             pop     di                ;   stack
             push    dx                ; save
             push    bx                ;  GPZ
             push    ax                ;   on stack
             mov     ax, 00088h        ; load
             mov     bx, 066A5h        ;  real constant
             mov     dx, 03429h        ;   1.801617224813e2
             push    di                ; save
             push    si                ;  z on
             push    cx                ;   stack
             call    RealAdd           ; add z

             pop     cx                ; get
             pop     si                ;  z from
             pop     di                ;   stack
             call    RealMulF          ; multiply result by z

             mov     cx, 0388Bh        ; load
             mov     si, 0C182h        ;  real constant
             mov     di, 05233h        ;   1.681617371664e3
             call    RealAdd           ; compute QZ

             pop     cx                ; get
             pop     si                ;  GPZ from
             pop     di                ;   stack
             push    di                ; save
             push    si                ;  GPZ on
             push    cx                ;   stack
             call    RealSub           ; compute QZ - GPZ

             pop     cx                ; get
             pop     si                ;  GPZ from
             pop     di                ;   stack
             call    RealdivRev        ; compute GPZ / (QZ-GPZ)

             mov     cx, 80h           ; load
             xor     si, si            ;  real constant
             mov     di, si            ;   0.5
             call    RealAdd           ; compute 0.5 + GPZ / (QZ-GPZ)

             pop     cx                ; get n
             inc     cx                ; compute n+1
             add     al, cl            ; adjust exponent by adding n+1
             pop     bp                ; restore caller's frame pointer
             jz      $exp_overfl       ; overflow error if exponent overflows

             ret                       ; done

$exp_overfl: jmp     ROverflow         ; error 205, floating point overflow

PowECst      dw      00080h, 017F7h, 0B172h ; ln(2)-eps = 6.931471803691e-1
             dw      00060h, 079ACh, 0D1CFh ;       eps = 1.908214929385e-10
             dw      05C81h, 03B29h, 038AAh ; 1/ln(2)   = 1.442695040889e+0

RExp         ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END