; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Frac Function                              *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FFRC

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  RFrac
             PUBLIC  RealFrac

;-------------------------------------------------------------------------------
; RFrac represents the standard function Frac. It computes the fractional part
; of a TURBO Pascal six byte floating point number. This routine is realized as
; a selfcontained routine rather than as a combination of the RInt and RealSub
; routines.
;
; INPUT:     DX:BX:AX  floating point number
;
; OUTPUT:    DX:BX:AX  fractional part of floating point number
;
; DESTROYS:  AX,BX,CX,DX,Flags
;-------------------------------------------------------------------------------

RFrac        PROC    FAR

RealFrac:    cmp     al, 80h           ; is number < 1 ?
             jbe     $unchanged        ; yes, that is the result

             cmp     al, 0a8h          ; is number > 2^39 ?
             ja      $frac_zero        ; yes, no fractional part

             mov     ch, 7fh           ; generate mask for sign bit
             or      ch, dh            ; get sign bit
             push    cx                ; save sign mask
             jmp     $shift_start      ; start left shift

             nop                       ; filler

$frac_shift8:sub     al, 8             ; adjust exponent
             mov     dh, dl            ; shift
             mov     dl, bh            ;  mantissa
             mov     bh, bl            ;   8 bits
             mov     bl, ah            ;    to the
             xor     ah, ah            ;     left

$shift_start:cmp     al, 88h           ; another byte shift possible ?
             ja      $frac_shift8      ; yes, do it

             INCLUDE ALIGN.ASM

$frac_shift1:dec     ax                ; adjust exponent
             add     ah, ah            ; shift
             adc     bx, bx            ;  mantissa
             adc     dx, dx            ;   1 bit to the left
             cmp     al, 80h           ; another bit shift necessary ?
             ja      $frac_shift1      ; yes, do it

             mov     cx, dx            ; test if
             or      ch, ah            ;   resulting
             or      cx, bx            ;    mantissa is zero
             pop     cx                ; get back sign mask
             jz      $frac_zero        ; yes, return zero

$frac_norm:  test    dh, dh            ; mantissa normalized ?
             js      $frac_exit        ; yes

             add     ah, ah            ; shift
             adc     bx, bx            ;  mantissa
             adc     dx, dx            ;   1 bit to the left
             dec     al                ; adjust exponent
             jnz     $frac_norm        ; if no underflow, cont. normalization

$frac_zero:  xor     ax, ax            ; load
             xor     bx, bx            ;  a
             xor     dx, dx            ;   zero

$frac_exit:  and     dh, ch            ; mask out sign bit if necessary
$unchanged:  ret

RFrac        ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END