; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Int Function                               *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FINT

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  RInt

;-------------------------------------------------------------------------------
; RInt represents the standard function Int. It computes the integral part of a
; TURBO-Pascal six byte floating point number, the result being a floating point
; number.
;
; INPUT:     DX:BX:AX  floating point number
;
; OUTPUT:    DX:BX:AX  integral part of floating point number
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RInt         PROC    FAR

             cmp     al, 0a8h          ; is argument > 2^39 ?
             jnb     $no_change        ; yes, return number unchanged

             cmp     al, 80h           ; argument < 1 ?
             jbe     $res_zero         ; yes, return zero

             mov     cx, ax            ; save
             mov     si, bx            ;  original
             mov     di, dx            ;   argument
             cmp     al, 88h           ; argument >= 2^7 ?
             sbb     dh, dh            ; yes, DH=0 (else DH=FFh)
             cmp     al, 90h           ; argument >= 2^15 ?
             sbb     dl, dl            ; yes, DL=0 (else DL=FFh)
             cmp     al, 98h           ; argument >= 2^23 ?
             sbb     bh, bh            ; yes, BH=0 (else BH=FFh)
             cmp     al, 0a0h          ; argument >= 2^31 ?
             sbb     bl, bl            ; yes, BL=0 (else BL=FFh)
             not     dx                ; generate first
             not     bx                ;  32 bits of mask
             and     ax, 7             ; clear LSB of mask, test if 1-bit shift
             jz      $shiftr_done      ; no further shifting required

             INCLUDE ALIGN.ASM

$shift_loop: add     al, 0ffh          ; dec. shift counter, set carry flag
             rcr     dx, 1             ; extend mask 1 bit
             rcr     bx, 1             ;  to the
             rcr     ah, 1             ;   right
             jnz     $shift_loop       ; shift until counter zero

$shiftr_done:and     dx, di            ; mask out mantissa
             and     bx, si            ;  bits containing
             and     ch, ah            ;   integral part of number
             xchg    ax, cx            ; get back exponent
             ret                       ; done

$res_zero:   xor     ax, ax            ; load
             xor     bx, bx            ;   a
             cwd                       ;    zero

$no_change:  ret

RInt         ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END