; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Logarithm                                  *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FLOG

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   CmpMantissa  : NEAR
             EXTRN   HaltError    : NEAR
             EXTRN   RealAdd      : NEAR
             EXTRN   RealDivRev   : NEAR
             EXTRN   RealFloat    : NEAR
             EXTRN   RealMulFNochk: Near
             EXTRN   RealMulNoChk : NEAR
             EXTRN   RealPoly     : NEAR
             EXTRN   RealSub      : NEAR
             EXTRN   ROverflow    : NEAR
             EXTRN   ShortMulRev  : NEAR
; Publics

             PUBLIC  RLn

;-------------------------------------------------------------------------------
; RLn computes the natural logarithm of its argument. It uses a polynomial
; approximation to compute the natural logarithm of the reduced argument z. The
; reduced argument satisfies the inequality |z| <= (sqrt(2)-1)^2. RLog10 and
; RLog2 are additional routines that compute the logarithms base two and ten,
; respectively. Both first execute RLn to compute the natural logarithm and
; then proceed to multiply the result with the appropriate constants to get
; Log10 and Log2. The following polynomial approximation is used to compute
; the natural logarithm:
;
; rz := ((((0.09790802001953*z^2 + 0.1108818338371)*z^2 + 0.1428605246897)*z^2
;           0.1999999783036)*z^2 + 0.3333333333786)*z^2 * z + z
;
; This approximation has a theoretical maximum relative error of 3.20e-14.
; Maximum observed error when evaluated in REAL arithmetic is 9.31e-13.
;
; If the argument is negative or zero, runtime error 207 is invoked through the
; error handler.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  ln, log10, log2 of argument depending on routine called
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RLn          PROC    FAR

             mov     di, offset $log_end; push address of ln tail-routine

$start_log:  test    dh, dh             ; x negative ?
             js      $range_err         ; yes, error

             test    al, al             ; x zero ?
             jz      $range_err         ; yes, error

             push    di                 ; save log routine tail address
             mov     cx, 0fa81h         ; CL = exponent of constant a = 1,
             mov     si, 0f333h         ;  DI:SI:CH = mantissa
             mov     di, 03504h         ;   of 0.5*sqrt(2)
             call    CmpMantissa        ; compare mantissas of x and 0.5*sqrt(2)

             jnc     $gt_root2          ; if mantissa x > mantissa 0.5*sqrt(2)

             dec     cx                 ; exponent of constant a = 0.5
             dec     ax                 ; exponent = exponent - 1

$gt_root2:   push    ax                 ; save exponent of x
             mov     al, 80h            ; x = mantissa of x
             xor     ch, ch             ; clear LSB of constant a
             push    cx                 ; save exponent of constant a
             xor     si, si             ; real constant
             mov     di, si             ;  a = 1 or a = 0.5
             call    RealSub            ; x-a

             pop     cx                 ; get exponent of constant a
             push    dx                 ; save
             push    bx                 ;  x-a
             push    ax                 ;   on stack
             inc     cx                 ; create
             xor     si, si             ;  constant
             mov     di, si             ;   2a
             call    RealAdd            ; compute (x-a) + 2a = x+a

             pop     cx                 ; get
             pop     si                 ;  back
             pop     di                 ;   x-a
             call    RealDivRev         ; compute (x-a)/(x+a)

             mov     cx, 5              ; polynomial has five coefficients
             mov     di, offset LN_COEFF; pointer to first coefficient
             xor     si, si             ; polynomial of type P(x^2)*x+x
             call    RealPoly           ; z+z*p(z^2), max. rel. err. 2.6e-12

             add     al, 0ffh           ; compute rz := 2 * (z + z * p(^2))
             adc     al, 1              ;  except when result is zero
             pop     cx                 ; get exponent
             push    dx                 ; save
             push    bx                 ;  rz on
             push    ax                 ;   stack
             xchg    ax, cx             ; AL = exponent
             sub     al, 80h            ; compute n = exponent - $80
             cbw                        ; convert n to word
             cwd                        ; convert n to longint
             call    RealFloat          ; compute float (n)

             mov     cx, 0d280h         ; load
             mov     si, 017f7h         ;  real constant
             mov     di, 03172h         ;   ln(2)
             call    ShortMulRev        ; compute n*ln(2),max. rel. err. 1.12e-12

             pop     cx                 ; get
             pop     si                 ;  rz from
             pop     di                 ;   stack
             jmp     RealAdd            ; compute rz + n * ln(2)

$range_err:  mov     ax, 207            ; load error code 207
             jmp     HaltError          ; execute error handler

             INCLUDE ALIGN.ASM

$log_end:    ret

LN_COEFF     db      07dh,               084h,048h ; 9.790802001953e-2
             db      07dh,068h,0d0h,003h,016h,063h ; 1.108818338371e-1
             db      07eh,0bah,085h,007h,04ah,012h ; 1.428605246897e-1
             db      07eh,00fh,058h,0cbh,0cch,04ch ; 1.999999783036e-1
             db      07fh,00eh,0abh,0aah,0aah,02ah ; 3.333333333786e-1

RLn          ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END