; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Square Root                                *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FSRT

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   HaltError: NEAR

; Publics

             PUBLIC  RSqrt

;-------------------------------------------------------------------------------
; RSqrt computes the square root of it argument. For a negative argument,
; runtime error 207 is invoked. The routine uses a estimate-and-correct method
; similar to that used in the division routine. This algorithm is based on the
; longhand method taught in school. Since there can be no tie case in rounding
; when computing the square root, no guard and sticky flags are needed to round
; to nearest or even in compliance with the IEEE floating point standard.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  square root of argument
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

ROOT_EXT     PROC   FAR

$zero_sqrt:  ret                     ; return argument

$sqrt_err:   mov     ax, 207           ; load error code
             jmp     HaltError         ; execute error handler

ROOT_EXT     ENDP

             INCLUDE ALIGN.ASM

RSqrt        PROC    FAR

             test    al, al            ; argument zero ?
             jz      $zero_sqrt        ; yes, return zero

             test    dh, dh            ; argument negative ?
             js      $sqrt_err         ; yes, error

             push    bp                ; save TURBO-Pascal frame pointer

             shr     al, 1             ; divide biased exponent by 2
             setc    cl                ; CL = 0, if expo even, else CL = -1
             adc     al, 40h           ; make new biased exponent
             push    ax                ; save new exponent
             xor     al, al            ; clear LSB of a3
             or      dh, 80h           ; set hidden bit
             shr     dx, cl            ; divide argument
             rcr     bx, cl            ;  by 2 if
             rcr     ax, cl            ;   expo odd
             xchg    ax, cx            ; argument in DX:BX:CX

             mov     si, dx            ; save a1
             mov     di, bx            ; save a2
             mov     bp, cx            ; save a3
             mov     bh, dh            ; get MSB of a1

             stc                       ; generate 4 bit approximation
             rcr     bh, 1             ;  in hi-nibble of BH
             neg     al                ; adjust approximation (subtract 20)
             and     al, 20            ;  for arguments
             sub     bh, al            ;   between 4000 0000 00 and 7FFF FFFF FF

             mov     al, 0ffh          ; default quotient = FFh
             cmp     dh, bh            ; division overflow ?
             jnb     $no_div0          ; yes, use default quotient

             mov     ax, dx            ; get a1
             div     bh                ; compute a1 / approximation

$no_div0:    add     bh, al            ; add result to approximation
             rcr     bh, 1             ;  and divide by 2 yields 8 bit approx.
             mov     bl, 0ffh          ; BX >= Trunc(Sqrt(a1))
             mov     ax, 0ffffh        ; default quotient
             cmp     dx, bx            ; division overflow ?
             jnb     $no_div1          ; yes, use default quotient

             mov     ax, di            ; get a2
             div     bx                ; compute a1:a2 / approximation

$no_div1:    add     ax, bx            ; add result to approximation
             rcr     ax, 1             ;  and divide by 2 yields 16 bit approx.
             mov     bx, ax            ; save q1
             mul     ax                ; DX:AX = q1*q1
             sub     di, ax            ; compute
             sbb     si, dx            ;  remainder (SI:DI)
             jnc     $quot1_ok         ; remainder must be positive

             INCLUDE ALIGN.ASM

$corr_quot1: dec     bx                ; try next smaller quotient q1
             stc                       ; adjust
             adc     di, bx            ;  remainder
             adc     si, 0             ;   to new
             add     di, bx            ;     value
             adc     si, 0             ;      of quotient
             js      $corr_quot1       ; until remainder becomes positive

$quot1_ok:   xchg    ax, cx            ; concat rest of argument to remainder
             mov     dx, di            ; get remainder lo-word
             shr     si, 1             ; divide remainder (from here SI=0 !!!)
             rcr     dx, 1             ;  by two so it fits into 32 bits
             rcr     ax, 1             ; DX:AX = remainder / 2
             mov     cx, 0ffffh        ; load default quotient
             cmp     dx, bx            ; division overflow ?
             jnb     $no_div2          ; yes, skip division and use default quot

             div     bx                ; AX = q2
             xchg    ax, cx            ; BX:CX = q1:q2

$no_div2:    mov     ax, cx            ; get q2
             mul     bx                ; q1*q2
             sub     bp, ax            ; subtract product
             sbb     di, dx            ;  from
             sub     bp, ax            ;   old
             sbb     di, dx            ;    remainder
             mov     ax, cx            ; get q2
             mul     ax                ; q2*q2
             neg     ax                ; compute
             sbb     bp, dx            ;   new
             sbb     di, si            ;    remainder (DI:BP:AX)
             jns     $quot2_ok         ; remainder must be positive

$corr_quot2: dec     cx                ; try next smaller value for q2
             stc                       ; adjust
             adc     ax, cx            ;  value
             adc     bp, bx            ;   of remainder
             adc     di, si            ;    according
             add     ax, cx            ;     to changed
             adc     bp, bx            ;      value
             adc     di, si            ;       of q2
             js      $corr_quot2       ; until remainder positive

$quot2_ok:   mov     si, ax            ; DI:BP:SI = remainder, BX:CX = quot
             mov     dx, bp            ; divide
             shr     di, 1             ;  remainder by two and
             rcr     dx, 1             ;   stuff 32 most significant bits
             rcr     ax, 1             ;    into DX:AX
             mov     di, 0ffffh        ; load default quotient q3
             cmp     dx, bx            ; division overflow ?
             jnb     $no_div3          ; yes, use default quotient and skip div.

             div     bx                ; AX = q3
             xchg    ax, di            ; BX:CX:DI = q1:q2:q3, BP:SI:0:0 = rem

$no_div3:    mov     ax, di            ; get q3
             mul     bx                ; q1*q3
             sub     si, ax            ; subtract
             sbb     bp, dx            ;  2*q1*q3
             sub     si, ax            ;   from
             sbb     bp, dx            ;    remainder
             mov     ax, di            ; get q3
             mul     cx                ; q2*q3
             push    bx                ; save q1
             xor     bx, bx            ; BP:SI:BX:0 = remainder
             sub     bx, ax            ; subtract
             sbb     si, dx            ;  2*q2*q3
             sbb     bp, 0             ;   from
             sub     bx, ax            ;    remainder
             sbb     si, dx            ;     "
             sbb     bp, 0             ;      "
             mov     ax, di            ; get q3
             mul     ax                ; q3*q3
             neg     ax                ; subtract
             sbb     bx, dx            ;  q3*q3
             sbb     si, 0             ;   from remainder
             sbb     bp, 0             ; BP:SI:BX:AX = remainder
             pop     dx                ; DX:CX:DI = q1:q2:q3
             jns     $quot3_ok         ; remainder must be positive

$corr_quot3: dec     di                ; try next smaller value for q3
             stc                       ; adjust
             adc     ax, di            ;  value
             adc     bx, cx            ;   of
             adc     si, dx            ;    remainder
             adc     bp, 0             ;     according
             add     ax, di            ;      to
             adc     bx, cx            ;       changed
             adc     si, dx            ;        value of
             adc     bp, 0             ;         quotient
             js      $corr_quot3       ; until remainder positive

$quot3_ok:   xchg    ax, di            ; get result
             mov     bx, cx            ;  into DX:BX:AX
             add     ax, 80h           ; round result
             adc     bx, 0             ;  (no tie case and
             adc     dx, 0             ;   no mantissa overflow possible)
             pop     cx                ; get saved exponent
             mov     al, cl            ; stuff exponent into result
             and     dh, 7fh           ; clear sign bit

             pop     bp                ; restore TURBO-Pascal frame pointer
             ret

RSqrt        ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END